#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)

import os
import pandas as pd
import numpy as np

from bokeh.plotting import figure
from bokeh.layouts import gridplot, column
from bokeh.io import save, output_file
from bokeh.models import ColumnDataSource, HoverTool, Span

from dqc_wrapper.utils.utils_files import *
from dqc_wrapper.utils.utils_grib import *

'''
Class: Bokeh plots
---
- General plot functions for histograms 
- Statistical formulas needed to produce the plots
'''
class bokehPlot:

    def __init__(self, outfname, csvs, csvh, vv, vmin, vmax, mm):
        self.outfile_hist = outfname
        self.outfile_stat = outfname.replace('.html', '_boxplot.html')
        self.csv_hist = csvh
        self.csv_stat = csvs
        self.valid_min = vmin
        self.valid_max = vmax
        self.varname = vv
        self.metadata = mm
        self.regions = {
          'total' : '[all latitudes]',
          'N_polar' : '[60 to 90 deg]', 
          'N_midlat' : '[30 to 60 deg]', 
          'N_hadley' : '[0 to 30 deg]', 
          'S_hadley' : '[0 to -30 deg]', 
          'S_midlat' : '[-30 to -60 deg]', 
          'S_polar' : '[-60 to -90 deg]'
        }

    '''
    Plot a boxplot diagram
    '''
    def plot_boxplot(self, col='File', pw=1000, ph=550):

      # generate some synthetic time series for six different categories
      df = pd.read_csv(self.csv_stat)
      tt = get_dates_from_files(df[col])
      if len(tt) > 2:
        tdiff = np.diff(tt)[0] * 0.5
      else:
        tdiff = 10 

      # Define figure
      pw=1000
      ph=600

      # Set up the figure same as before
      p = figure(plot_width = pw, plot_height = ph,
                 title = 'Observed values boxplot',
                 x_axis_label = 'Time', 
                 y_axis_label = self.varname,
                 x_axis_type='datetime',
                 tools='pan,wheel_zoom,box_zoom,reset')
      p.xaxis.axis_label_text_font_size = '8pt'

      for ii, r in df.iterrows():

          upper = r['q3']+1.5*r['IQD']
          lower = r['q1']-1.5*r['IQD']

          # stems
          p.segment(tt[ii], upper, tt[ii], r['q3'], line_color='black', legend_label='-/+ 1.5*IQD')
          p.segment(tt[ii], lower, tt[ii], r['q1'], line_color='black', legend_label='-/+ 1.5*IQD')

          # boxes
          p.vbar(x=tt[ii], bottom=r['Med'], top=r['q3'], width=tdiff, fill_color='#E08E79', line_color='black', legend_label='Q3 (75%)')
          p.vbar(x=tt[ii], bottom=r['q1'], top=r['Med'], width=tdiff, fill_color='#3B8686', line_color='black', legend_label='Q1 (25%)')

          # whiskers (almost-0 height rects simpler than segments)
          p.rect(tt[ii], lower, tdiff*0.5, 0.01, line_color='black')
          p.rect(tt[ii], upper, tdiff*0.5, 0.01, line_color='black')

      # Min/Max lines
      p.line(tt, df['Min'], color='green', legend_label='Observed min')
      p.line(tt, df['Max'], color='red', legend_label='Observed max')

      return p

    '''
    Plot a simplified boxplot diagram (only lines)
    '''
    def plot_boxplot_lines(self, col='File', pw=1000, ph=500):

      # generate some synthetic time series for six different categories
      df = pd.read_csv(self.csv_stat)
      tt = get_dates_from_files(df[col])

      # Define figure
      pw=1000
      ph=600

      # Set up the figure same as before
      p = figure(plot_width = pw, plot_height = ph,
                 title = '{} variability [all latitudes, all timesteps]'.format(self.metadata['varname']),
                 x_axis_label = 'Time', 
                 y_axis_label = '{} ({})'.format(self.varname, self.metadata['unit']),
                 x_axis_type='datetime',
                 toolbar_location='above',
                 tools='pan,wheel_zoom,box_zoom,reset')
      p.toolbar.logo = None

      # whiskers / IQD
      p.line(tt, df['q1']-1.5*df['IQD'], color='#ED2939', line_dash=[1, 2])
      p.line(tt, df['q3']+1.5*df['IQD'], color='#ED2939', line_dash=[1, 2])
      p.varea(x=tt, y1=df['q1']-1.5*df['IQD'], y2=df['q3']+1.5*df['IQD'], color='#A45A52', alpha=0.1, legend_label='1.5*IQD (Inter-quantile dev)')

      # Boxes (Q1-Q3)
      p.line(tt, df['q1'], color='#800000', line_dash=[1, 2])
      p.line(tt, df['q3'], color='#800000', line_dash=[1, 2])
      p.varea(x=tt, y1=df['q1'], y2=df['q3'], color='#800000', alpha=0.25, legend_label='Quantiles (25% to 75%)')

      # Min/Max lines
      p.line(tt, df['Max'], color='#8D021F')
      p.circle(tt, df['Max'], color='#8D021F', legend_label='Maximum')
      p.line(tt, df['Med'], color='#800000', legend_label='Median')
      p.line(tt, df['Min'], color='#FF2800')
      p.circle(tt, df['Min'], color='#FF2800', legend_label='Minimum')
      
      return p

    '''
    Add valid min/max to plot (vertical lines + highlights)
    '''
    def add_valid_minmax(self, fig, l, y, nz):

        vert_lines = []

        if not(self.valid_max is None):
          vert_lines.append(Span(location=self.valid_max, dimension='height', line_color='red', line_dash='dashed', line_width=1.5))       
          vmax = np.where(l[nz] > self.valid_max)
          if len(vmax):
            fig.circle((l[nz])[vmax], y[vmax], color='red', fill_color='white', legend_label='outliers')

        if not(self.valid_min is None):
          vert_lines.append(Span(location=self.valid_min, dimension='height', line_color='red', line_dash='dashed', line_width=1.5))        
          vmins = np.where(l[nz] < self.valid_min)
          if len(vmins):
            fig.circle((l[nz])[vmins], y[vmins], color='red', fill_color='white', legend_label='outliers')
        
        # Add vertical lines
        if vert_lines != []:
          fig.renderers.extend(vert_lines)

    '''
    Plot a single histogram, return figure
    '''
    def plot_histogram_figure(self, df, tag='total', col='File', pw=500, ph=250):

        # Get bins (left and right) + remove first element
        l = np.delete(df.columns.values, 0).astype(float) 
        xdist = abs(l[0]-l[1])
        r = [ l+xdist for v in l ]
        
        # Get count + remove first element
        y = np.delete(df.loc[df[col] == tag].values, 0).astype(float)

        # Get non-zero bins
        nz = np.nonzero(y)

        # Set up the figure same as before
        fig = figure(plot_width = pw, plot_height = ph,
                   title = '{} distribution {}'.format(self.metadata['varname'], self.regions[tag]),
                   x_axis_label = '{} ({})'.format(self.varname, self.metadata['unit']),
                   y_axis_label = 'percent (%)',
                   toolbar_location='above',
                   tools='pan,wheel_zoom,box_zoom,reset')

        # xaxis sizes
        if tag != 'total':
          fig.xaxis.axis_label_text_font_size = '8pt'
          fig.yaxis.axis_label_text_font_size = '8pt'
          fig.title.text_font_size = '8pt'

        # Percentages instead of number of bins
        perc = 100.0 * y[nz]/max(y[nz])

        # Histograms and points where we have values
        fig.vbar(x=l[nz], width=xdist, bottom=0, top=perc, color='#941333')

        # Dash where there are values
        fig.dash(x=l[nz], y=perc, color='#941333')
        fig.toolbar.logo = None
        
        # Vertical lines indicating valid_min/valid_max if available
        self.add_valid_minmax(fig, l, perc, nz)
        fig.legend.visible = (tag=='total')

        return fig

    '''
    Plot multiple histograms into a grid and return a grid of the plots (file) 
    '''
    def plot_histograms(self):

        # Read histograms into pandas
        df_hist = pd.read_csv(self.csv_hist)

        # Total histogram
        p0 = self.plot_histogram_figure(df_hist, tag='total', pw=1000, ph=500)

        # Latitude regions
        p1 = None
        try:
          figs = []
          for t in [ 'N_polar', 'S_polar', 'N_midlat', 'S_midlat', 'N_hadley', 'S_hadley' ]:
            figs.append(self.plot_histogram_figure(df_hist, tag=t))

          # Group plots into grid
          p1 = gridplot(figs, ncols=2, toolbar_options=dict(logo=None))
        except:
          pass # Dataset has no regions

        # Boxplot dates
        p2 = self.plot_boxplot_lines()

        # Save to file
        print('INFO: Saving plot {}'.format(self.outfile_hist))
        output_file(self.outfile_hist)
        
        if p1 is None or p2 is None:
          save(p0)
        else:
          save(column(p0, p1, p2))
